/*
 * gabien-app-ppj - Editor/Player for 'PiyoPiyo' music files
 * Written starting in 2015 by contributors (see CREDITS.txt)
 * To the extent possible under law, the author(s) have dedicated all copyright and related and neighboring rights to this software to the public domain worldwide. This software is distributed without any warranty.
 * You should have received a copy of the CC0 Public Domain Dedication along with this software. If not, see <http://creativecommons.org/publicdomain/zero/1.0/>.
 */

package libpiyo;

import gabien.uslx.audio.ISample;
import gabien.uslx.audio.ISample.ArrayF32;
import gabien.uslx.audio.ISample.Lerped;
import gabien.uslx.audio.ISample.Looped;

/**
 * Created on 09/12/15.
 */
public class SimpleMixer {
    public Channel[] channels;
    private boolean lerp;
    // Basic volume adjustment algorithm which should produce passable results.
    private boolean preventClipping;
    private long maxVolLast;

    public SimpleMixer(int channelCount, boolean preventClip, boolean doLerp) {
        channels = new Channel[channelCount];
        preventClipping = preventClip;
        lerp = doLerp;
    }

    public class Channel {
        private ISample data;
        public double pos = 0;
        public double volumeLeft = 0, volumeRight = 0, pitch = 1;

        private Channel() {

        }

        private short get() {
            float res = data.get(pos);
            pos += pitch;
            if (res < 0)
                return (short) (res * 32768);
            return (short) (res * 32767);
        }
        
        private int[] createData(int amount) {
            int[] s = new int[amount * 2];
            for (int px = 0; px < amount; px++) {
                short V = get();
                s[px * 2] = (int) (V * volumeLeft);
                s[(px * 2) + 1] = (int) (V * volumeRight);
            }
            return s;
        }

        public void playSound(double p, double VolL, double VolR, short[] sound, boolean isLooping) {
            float[] dt = new float[sound.length];
            for (int i = 0; i < sound.length; i++) {
                int snd = sound[i];
                dt[i] = (snd < 0) ? (snd / 32768f) : (snd / 32767f);
            }
            ISample base = new ISample.ArrayF32(dt);
            if (isLooping)
                base = new ISample.Looped(base, 0, sound.length);
            playSound(p, VolL, VolR, base);
            pos = 0;
        }

        public void playSound(double p, double VolL, double VolR, ISample sound) {
            volumeLeft = VolL;
            volumeRight = VolR;
            // Sanity check
            if (p <= 0)
                p = 1;
            if (p == Double.POSITIVE_INFINITY)
                p = 1;
            if (p == Double.NEGATIVE_INFINITY)
                p = -1;
            pitch = p;
            data = lerp ? new ISample.Lerped(sound) : sound;
            pos = 0;
        }

        public void setVolume(double d, double e) {
            volumeLeft = d;
            volumeRight = e;
        }
    }

    /**
     * Allocates a channel slot and returns the resulting channel.
     */
    public Channel createChannel() {
        Channel jc = new Channel();
        for (int p = 0; p < channels.length; p++) {
            if (channels[p] == null) {
                channels[p] = jc;
                break;
            }
        }
        return jc;
    }

    public short[] pullData(int amount) {
        short[] data = new short[amount * 2];
        int[] totalLeft = new int[amount];
        int[] totalRight = new int[amount];
        for (Channel js : channels) {
            if (js == null)
                continue;
            int[] r = js.createData(amount);
            for (int px = 0; px < amount; px++) {
                totalLeft[px] += r[(px * 2)];
                totalRight[px] += r[(px * 2) + 1];
            }
        }

        // Volume adjustment.
        long maxVol = 32768;
        if (preventClipping) {
            for (int px = 0; px < amount; px++) {
                maxVol = Math.max(maxVol, Math.abs(totalLeft[px]));
                maxVol = Math.max(maxVol, Math.abs(totalRight[px]));
            }
            if (maxVolLast < maxVol)
                maxVolLast = maxVol;
            maxVol = maxVolLast;
            // Sample-rate-dependent, but it's assumed to be @ 22050hz where this algorithm will return to normal in 1.5s or so.
            maxVolLast -= amount;
        }

        for (int px = 0; px < amount; px++) {

            long l = ((totalLeft[px] * 32768L) / maxVol);
            long r = ((totalRight[px] * 32768L) / maxVol);

            if (l > 32767)
                l = 32767;
            if (l < -32768)
                l = -32768;
            if (r > 32767)
                r = 32767;
            if (r < -32768)
                r = -32768;
            data[px * 2] = (short) l;
            data[(px * 2) + 1] = (short) r;
        }
        return data;
    }

    public void deleteChannel(Channel ic) {
        for (int p = 0; p < channels.length; p++) {
            if (channels[p] == ic)
                channels[p] = null;
        }
    }
}
